/*
* pixelserv.c a small mod to public domain
** server.c -- a stream socket server demo
* from http://beej.us/guide/bgnet/
* single pixel http string from
* http://proxytunnel.sourceforge.net/pixelserv.php
* V1 Proof of concept mstombs www.linkysinfo.org 06/09/09
* V2 usleep after send to delay socket close 08/09/09
* V3 TCP_NODELAY not usleep 09/09/09
* V4 daemonize with syslog 10/09/09
* V5 usleep back in 10/09/09
* V6 only use IPV4, add linger and shutdown to avoid need for sleep 11/09/09
* Consistent exit codes and version stamp
* V7 use shutdown/read/shutdown to cleanly flush and close connection
*/

#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>
#include <errno.h>
#include <string.h>
#include <sys/types.h>
#include <sys/socket.h>
#include <netinet/in.h>
#include <netdb.h>
#include <arpa/inet.h>
#include <sys/wait.h>
#include <signal.h>
#include <syslog.h>

#define PORT "80"  // the port users will be connecting to

#define BACKLOG 10	 // how many pending connections queue will hold

void sigchld_handler(int s)
{
	while(waitpid(-1, NULL, WNOHANG) > 0);
}

/* Handler for the SIGTERM signal (kill) */
static void
sigterm(int sig)
{
	signal(sig, SIG_IGN);	/* Ignore this signal while we are quitting */
	syslog(LOG_NOTICE, "received signal SIGTERM, exiting");
	exit(1);
}

#ifdef TEST
// get sockaddr, IPv4 or IPv6:
void *get_in_addr(struct sockaddr *sa)
{
	if (sa->sa_family == AF_INET) {
		return &(((struct sockaddr_in*)sa)->sin_addr);
	}

	return &(((struct sockaddr_in6*)sa)->sin6_addr);
}
#endif

int main(void)
{
	int sockfd, new_fd;  // listen on sock_fd, new connection on new_fd
	struct addrinfo hints, *servinfo, *p;
	struct sockaddr_storage their_addr; // connector's address information
	socklen_t sin_size;
	struct sigaction sa;
	int yes=1;
#ifdef TEST
	char s[INET6_ADDRSTRLEN];
#endif
	int rv;

/*	static const char http_head[] =
		"HTTP/1.1 200 OK\r\n"
		"Content-type: image/gif\r\n"
		"Accept-ranges: bytes\r\n"
		"Content-length: 43\r\n"
		"\r\n"; */
/* converted to numeric with */
/*	int i;
	for ( i = 0; i < sizeof http_head; i++)
	{
		printf("%u,", (int) http_head[i]);
	}
*/
	static const unsigned char httpnullpixel[] =    
		{ 72,84,84,80,47,49,46,49,32,50,48,48,32,79,75,13,10,67,111,110,
		  116,101,110,116,45,116,121,112,101,58,32,105,109,97,103,101,
          47,103,105,102,13,10,65,99,99,101,112,116,45,114,97,110,103,
          101,115,58,32,98,121,116,101,115,13,10,67,111,110,116,101,
          110,116,45,108,101,110,103,116,104,58,32,52,51,13,10,13,10,
/*		};
	static const unsigned char pixel[] = { */
		  71, 73, 70, 56, 57, 97, 1, 0, 1, 0, 128, 
		  0, 0, 255, 255, 255, 0, 0, 0, 33, 249, 4,
		  1, 0, 0, 0, 0, 44, 0, 0, 0, 0, 1, 0, 1, 0,
		  0, 2, 2, 68, 1, 0, 59 };

	printf( "pixelserv V7 compiled: %s from %s \n", __DATE__ " " __TIME__  , __FILE__ );

	if(daemon(0, 0)<0) {
		syslog(LOG_ERR, "failed to daemonize, quitting: %m");
		exit( EXIT_FAILURE );
	}

	openlog("pixerlserv", LOG_PID|LOG_CONS, LOG_DAEMON);

	memset(&hints, 0, sizeof hints);
	hints.ai_family = AF_INET; //AF_UNSPEC; /* AF_INET restricts to IPV4 */
	hints.ai_socktype = SOCK_STREAM;
	hints.ai_flags = AI_PASSIVE; // use my IP

	if ((rv = getaddrinfo(NULL, PORT, &hints, &servinfo)) != 0) {
		syslog(LOG_ERR, "getaddrinfo: %s", gai_strerror(rv));
		exit( EXIT_FAILURE );
	}

	// loop through all the results and bind to the first we can
	for(p = servinfo; p != NULL; p = p->ai_next) {
		if ((sockfd = socket(p->ai_family, p->ai_socktype,
				p->ai_protocol)) < 0) {
			syslog(LOG_WARNING, "socket: %m");
			continue;
		}

		if (setsockopt(sockfd, SOL_SOCKET, SO_REUSEADDR, &yes,
				sizeof(int)) < 0) {
			syslog(LOG_ERR, "setsockopt(SO_REUSEADDR): %m");
			exit( EXIT_FAILURE );
		}

		if (bind(sockfd, p->ai_addr, p->ai_addrlen) < 0) {
			close(sockfd);
			syslog(LOG_WARNING, "bind: %m");
			continue;
		}

		break;
	}

	if (p == NULL)  {
		syslog(LOG_ERR, "failed to bind, quitting");
		exit( EXIT_FAILURE );
	}

	freeaddrinfo(servinfo); // all done with this structure

	if (listen(sockfd, BACKLOG) < 0) {
		syslog(LOG_ERR, "listen: %m");
		exit( EXIT_FAILURE );
	}

	sa.sa_handler = sigchld_handler; // reap all dead processes
	sigemptyset(&sa.sa_mask);
	sa.sa_flags = SA_RESTART;
	if (sigaction(SIGCHLD, &sa, NULL) < 0) {
		syslog(LOG_ERR, "sigaction: %m");
		exit( EXIT_FAILURE );
	}

	/* set signal handler */
	memset(&sa, 0, sizeof(struct sigaction));
	sa.sa_handler = sigterm;

	if (sigaction(SIGTERM, &sa, NULL)) {
		syslog(LOG_ERR, "Failed to set SIGTERM handler. exiting");
		exit( EXIT_FAILURE );
	}

	syslog(LOG_NOTICE, "waiting for connections on port " PORT);

	while(1) {  // main accept() loop
		sin_size = sizeof their_addr;
		new_fd = accept(sockfd, (struct sockaddr *)&their_addr, &sin_size);
		if (new_fd < 0) {
			syslog(LOG_WARNING, "accept: %m");
			continue;
		}

		if (!fork()) { // this is the child process
			close(sockfd); // child doesn't need the listener
			rv = send(new_fd, httpnullpixel, sizeof httpnullpixel, 0);
			if (rv < 0) {
				syslog(LOG_WARNING, "send: %m");
			}
			/* clean way to flush read buffers and close connection */
			{   char buf[128];
    		if (shutdown(new_fd, SHUT_WR) == 0)
          		while( read(new_fd, buf, sizeof buf) > 0);
    		shutdown(new_fd, SHUT_RD);
			}
			close(new_fd);
			exit( EXIT_SUCCESS );
		}

		close(new_fd);  // parent doesn't need this
	}

	return EXIT_SUCCESS;
}

